// The strategy is to use TOPC_raw_begin/end_master_slave().
// Each slave precomputes quicksort() recursively down to PAR_LEVEL.
// The master then calls quicksort().  When it reaches PAR_LEVEL,
//   TOPC_raw_submit_task_input(_input_) is called within quicksort() by the
//   master, where _input_ is TOPC_MSG() containing indices:  {start,end}
// The slave then calls qs_do_task(), which calls:
//   quicksort(array,start,end,PAR_LEVEL);
// Another alternative would be for TOPC_MSG() to include the
//   entire subarray:  array[start..end]
//   which would avoid the need for the slave to precompute down to PAR_LEVEL,
//   but it would involve higher communication costs.

#ifndef PAR_LEVEL
#define PAR_LEVEL 3 /* Level at which master dispatches sorting to slave */
#endif
#ifndef PAR_TRACE
#define PAR_TRACE 1
#endif

#ifndef ARRAY_LEN
#define ARRAY_LEN 100
#endif

#include "topc.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>
#include <assert.h>

struct qs_task {
  int start;
  int end;
};

/*==================== MODIFICATION OF SEQUENTIAL CODE =================*/

/* main() will call this either with level == 0 or with level == PAR_LEVEL */
void quicksort( int array[], int start, int end, int level ) {
  int i = start, j = end, pivot = array[start];
  int tmp;
  static struct qs_task indices;

  level = level + 1;
  if ( start >= end ) return;

  if ( level == PAR_LEVEL ) {
    if ( TOPC_is_master() ) {
      indices.start = start;
      indices.end = end;
      TOPC_raw_submit_task_input( TOPC_MSG(&indices, sizeof(indices)) );
    }
    return;
  }

  /* interchange so nums <= pivot come first, and nums > pivot come later */
  while (1) {
    while( (array[i] <= pivot) && (i <= end) ) i++;
    while( array[j] > pivot ) j--; /* j>0 always, since pivot is an array val */
    if ( i < j ) {
      tmp = array[i];
      array[i] = array[j];
      array[j] = tmp;
    }
    else {
      i--;
      j++;
      array[start] = array[i];
      array[i] = pivot;
      break;
    }
  }

  assert( i == j-1 );
  quicksort( array, start, i - 1, level );  /* pivot==array[i] is in order */
  quicksort( array, j, end, level );
}

void print_array( char *string, int array[], int len ) {
  int i;
  printf("%s", string);
  for ( i=0; i < len; i++ )
    printf("%d ", array[i]);
  printf("\n");
}

/*==================== DEFINITION OF TOP-C CALLBACKS =================*/

int *array_global = NULL;
TOPC_BUF qs_do_task( void *input ) {
  struct qs_task * indices = input;
  quicksort( array_global, indices->start, indices->end, PAR_LEVEL );
  return TOPC_MSG( array_global + indices->start,
                   (indices->end - indices->start + 1) * sizeof(int) );
}
TOPC_ACTION qs_check_task_result( void *input, void *output ) {
  struct qs_task * indices = input;
  memcpy( array_global + indices->start, output,
          (indices->end - indices->start + 1) * sizeof(int) );
  return NO_ACTION;
}
void qs_trace_input(void *input) {
  printf("[%d, %d]", ((struct qs_task *)input)->start,
                     ((struct qs_task *)input)->end );
}
// PROBLEM:  CHANGE SYNTAX OF TOP-C FOR THIS:
void qs_trace_output2(void *input, void *output) {
  int len = ((struct qs_task *)input)->end
            - ((struct qs_task *)input)->start + 1;
  print_array( "", output, (len>8 ? 8 : len) );
}

/*==================== MAIN ROUTINE WITH INITIALIZATION ===================*/

int main( int argc, char **argv ) {
  int array[ARRAY_LEN];
  int i;

  TOPC_init(&argc, &argv, TOPC_OPTION("trace", PAR_TRACE), /* 1 is default */
                          TOPC_OPTION("trace_input", qs_trace_input),
                          TOPC_OPTION("trace_output2", qs_trace_output2) );

  // PROBLEM:  HOW TO ELEGANTLY DISTRIBUTE COMMON SEED?
  srandom( (unsigned int)time(NULL) );
  for ( i=0; i < ARRAY_LEN; i++ ) array[i] = ( random() & ((1<<10)-1) );
  array_global = array; /* pass array to qs_do_task & qs_check_task_result */

  /* Tell slaves to each do quicksort sequentially until PAR_LEVEL */
  if (!TOPC_is_master()) quicksort( array, 0, ARRAY_LEN - 1, 0 );

  /* The parallel stuff happens here */
  if (TOPC_is_master()) print_array("BEFORE:  ", array, ARRAY_LEN - 1);
  TOPC_raw_begin_master_slave( qs_do_task, qs_check_task_result,
                               NULL );  /* update_shared_memory not called */
  if (TOPC_is_master()) quicksort( array, 0, ARRAY_LEN - 1, 0 );
  TOPC_raw_end_master_slave();
  if (TOPC_is_master()) print_array("AFTER:  ", array, ARRAY_LEN - 1);

  TOPC_finalize();
  exit(0);
}
