 /**********************************************************************
  * TOP-C (Task Oriented Parallel C)                                   *
  * Copyright (c) 2000 Gene Cooperman <gene@ccs.neu.edu>               *
  *                    and Victor Grinberg <victor@ccs.neu.edu>        *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public         *
  * License as published by the Free Software Foundation; either       *
  * version 2.1 of the License, or (at your option) any later version. *
  *                                                                    *
  * This library is distributed in the hope that it will be useful,    *
  * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
  * Lesser General Public License for more details.                    *
  *                                                                    *
  * You should have received a copy of the GNU Lesser General Public   *
  * License along with this library (see file COPYING); if not, write  *
  * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
  * 330, Boston, MA 02111-1307 USA, or contact Gene Cooperman          *
  * <gene@ccs.neu.edu>.                                                *
  **********************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <stdlib.h> // malloc() for Alpha
#include <unistd.h> // alarm()
#include <signal.h> // signal()
#include <mpinu.h>  // MPINU implementation of MPI
#include "comm.h"

/***********************************************************************
 ***********************************************************************
 ** Communications functions, implemented over MPI
 ***********************************************************************
 ***********************************************************************/

/***********************************************************************
 * Initialize underlying communications library
 */
static void alarm_handler(int sig) {
  printf("TOP-C:  60 minutes elapsed without reply by master\n"
         "  slave exiting\n");
  exit(1);
}
void COMM_init(int *argc, char ***argv, struct opt_list *options_list) {
  int flag;
  MPI_Initialized(&flag);
  if (!flag) {
    MPI_Init(argc, argv);
    signal( SIGALRM, alarm_handler );
  }
}

/***********************************************************************
 * Shutdown underlying communications library
 */
void COMM_finalize() {
  int flag;
  MPI_Initialized(&flag);
  if (flag) MPI_Finalize();
}

/***********************************************************************
 * Test if underlying communications library is running
 */
TOPC_BOOL COMM_initialized() {
  int flag;
  MPI_Initialized(&flag);
  return flag;
}

/***********************************************************************
 * Return boolean, true if master and slave share memory
 */
int COMM_is_shared_memory() {
  return 0;  // false
}

/***********************************************************************
 * Get number of nodes
 */
int COMM_node_count() {
  int count;
  MPI_Comm_size(MPI_COMM_WORLD, &count);
  return count;
}

/***********************************************************************
 * Get rank of this node
 */
int COMM_rank() {
  int rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  return rank;
}

/***********************************************************************
 * Check if a message is instantly available (non-blocking)
 */
TOPC_BOOL COMM_probe_msg() {
  MPI_Status status;
  int flag;

  MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status);
  return flag;
}

/***********************************************************************
 * Send message
 */
void COMM_send_msg(void *msg, size_t msg_size, int dst, enum TAG tag) {
  MPI_Send(msg, (int)msg_size, MPI_BYTE, dst, tag, MPI_COMM_WORLD);
}

/***********************************************************************
 * Receive message into malloc'ed memory (blocking call)
 * receive from anybody, and set each of four param's if non-NULL
 */
void COMM_receive_msg(void **msg, size_t *msg_size, int *src, enum TAG *tag) {
  MPI_Status status;

  alarm(3600); // Allow one hour for master to reply, or we die.
  MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
  MPI_Get_count(&status, MPI_BYTE, (int *)msg_size);
  //if size_t 64 bits, only first 32 bits set
  if (msg_size) *msg_size = *(int *)msg_size;
  if (msg) {
    *msg                  = malloc(*msg_size);
    if (*msg == NULL && *msg_size > 0)
      { perror("No memory for receive buffer"); exit(1); }
  }
  MPI_Recv( (msg ? *msg : NULL), (int)*msg_size, MPI_BYTE,
             status.MPI_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
  if (src) *src            = status.MPI_SOURCE;
  if (tag) *tag            = status.MPI_TAG;

  alarm(0);  // Master replied.  Continue.
}

void COMM_free_receive_buf(void *msg) {
  free(msg); // balances malloc() in COMM_receive_msg()
}

/***********************************************************************
 ***********************************************************************
 * SMP-related functions (e.g.:  private (non-shared) variables)
 ***********************************************************************
 ***********************************************************************/

void *COMM_private_global(size_t size) { // trivial for non-shared memory
  static void *private_global = NULL;
  return ( private_global ? private_global : (private_global = malloc(size)) );
}
void COMM_begin_atomic_read() {};
void COMM_end_atomic_read() {};
void COMM_begin_atomic_write() {};
void COMM_end_atomic_write() {};
TOPC_BOOL COMM_is_in_atomic_section() { return 0; };
